(** Time-stamp: <modified the 30/01/2008 (at 09:44) by Erwan Jahier> *)

(*
Sous-module pour SyntaxTab 
*)
open Lxm
open SyntaxTree
open Errors

type 'a hereflaged =
  Here of 'a
| NotHere of Ident.long

type t = {
  st_consts: (Ident.t , (const_info srcflaged) hereflaged) Hashtbl.t ;
  st_types : (Ident.t , (type_info  srcflaged) hereflaged) Hashtbl.t ;
  st_opers : (Ident.t , (oper_info  srcflaged) hereflaged) Hashtbl.t ;
} 

(* Cration/initialisation d'une symbol_tab *)
let create () = {
   st_consts = Hashtbl.create 50;
   st_types  = Hashtbl.create 50;
   st_opers  = Hashtbl.create 50;
}

let find_type (this: t) (id: Ident.t) = (
  try 
    Hashtbl.find (this.st_types) id
  with Not_found -> 
    print_string ("*** unknown type "^(Ident.to_string id)^"\n");
    assert false
)
  
let find_const (this: t) (id: Ident.t) = (
  try 
    Hashtbl.find (this.st_consts) id
  with Not_found -> 
    print_string ("*** unknown const "^(Ident.to_string id)^"\n");
    assert false
)

let find_oper (this: t) (id: Ident.t) = (
  try
    Hashtbl.find (this.st_opers) id
  with Not_found -> 
    print_string ("*** unknown operator "^(Ident.to_string id)^"\n");
    assert false
)

(* Manip de SymbolTab.t *)
let add_import_const (this: t) (id: Ident.t) (aid: Ident.long) = (
  Hashtbl.replace (this.st_consts) id (NotHere aid)
)

let add_import_type (this: t) (id: Ident.t) (aid: Ident.long) = (
  Hashtbl.replace (this.st_types) id (NotHere aid)
)

let add_import_oper (this: t) (id: Ident.t) (aid: Ident.long) = (
  Hashtbl.replace (this.st_opers) id (NotHere aid)
)

let add_const (this: t) (n: Ident.t) (cix: const_info  srcflaged) = (
  Hashtbl.replace this.st_consts n (Here cix)
)

let add_type (this: t) (n: Ident.t) (tix: type_info srcflaged) = (
  Hashtbl.replace this.st_types n (Here tix) ;
  (* cas particulier des types enums *)
  match tix.it with
      EnumType (_, ecl) -> (
	let tname = Lxm.str tix.src in
	let treat_enum_const ec = (
	  let te = Named_type_exp { Ident.id_pack = None; Ident.id_id = tname} in
	  let tex = Lxm.flagit te tix.src in
	  let ci = EnumConst (ec.it, tex) in
	    add_const this ec.it (Lxm.flagit ci (ec.src))
	) in
	  List.iter treat_enum_const ecl
      )
    | _ -> ()
)

let add_oper (this: t) (n: Ident.t) (oix: oper_info  srcflaged) = (
	Hashtbl.add this.st_opers n (Here oix)
)

let iter_types this f = ( Hashtbl.iter f this.st_types)
let iter_consts this f = ( Hashtbl.iter f this.st_consts)
let iter_opers this f = ( Hashtbl.iter f this.st_opers)

let dump this = (
 (* A FAIRE (si besoin ...) *)
)
