(** Time-stamp: <modified the 30/01/2008 (at 09:53) by Erwan Jahier> *)

(**********************************************************
Sous-module pour SyntaxTab 
************************************************************

   Une SymbolTab.t est une association ``contextuelle'' entre
des ident simple (string) et ce  quoi ils correspondent.
   N.B. il est inutile de grer les idents absolus (pack+nom)
qui sont par dfinition NON-CONTEXTUELS.
   Dans une table, on a 3 espaces de noms, un par nature d'item
(const, type et oper).

***********************************************************)

type 'a hereflaged =
    Here of 'a            (* local items *) 
  | NotHere of Ident.long (* imported items *)

type t

val create : unit -> t

(* Manip de SymbolTab.t *)

(* Recherche d'items *)
val find_type : t -> Ident.t -> (SyntaxTree.type_info Lxm.srcflaged) hereflaged
val find_const : t -> Ident.t -> (SyntaxTree.const_info Lxm.srcflaged) hereflaged
val find_oper : t -> Ident.t -> (SyntaxTree.oper_info Lxm.srcflaged) hereflaged

(* Ajout de nom d'item imports (via uses) *)
val add_import_const : t -> Ident.t -> Ident.long -> unit
val add_import_type : t -> Ident.t -> Ident.long -> unit
val add_import_oper : t -> Ident.t -> Ident.long -> unit

(** Add local items declaration *)
val add_type : t -> Ident.t -> SyntaxTree.type_info  Lxm.srcflaged -> unit
val add_const : t -> Ident.t -> SyntaxTree.const_info  Lxm.srcflaged -> unit
val add_oper : t -> Ident.t -> SyntaxTree.oper_info  Lxm.srcflaged -> unit

(* Itrer sur les items *)

val iter_types : 
  t -> (Ident.t -> (SyntaxTree.type_info Lxm.srcflaged) hereflaged -> unit) -> unit
val iter_consts : 
  t -> (Ident.t -> (SyntaxTree.const_info Lxm.srcflaged) hereflaged -> unit) -> unit
val iter_opers : 
  t -> (Ident.t -> (SyntaxTree.oper_info Lxm.srcflaged) hereflaged -> unit) -> unit

val dump : t -> unit

