(** Time-stamp: <modified the 30/01/2008 (at 09:58) by Erwan Jahier> *)


(** (Raw) Abstract syntax tree of source programs. 

    This syntax tree represented by Hash tables.
*)

open Printf
open Lxm  (* pour la remonte au source *)
open Errors


(**********************************************)
(** Constructeur de type "avec erreur info"   *)
(**********************************************)
(* QUESTION: pourquoi ne pas le mettre dans le module Error? *)
type 'a error  = 
    Ok of 'a
  | Error of string


(*****************************************)
(* rsultat brut de l'analyse syntaxique *)
(*****************************************)
type t = 
    PRPackBody of packbody
  | PRPack_or_models of pack_or_model list
and
  pack_or_model =
    NSPack   of  pack_info srcflaged
  | NSModel  of  model_info srcflaged
and
  model_info = {
    mo_name  : Ident.pack_name ;
    mo_uses  : Ident.pack_name srcflaged list ;
    mo_needs : static_param srcflaged list ;
    (* N.B. CAS PARTICULIER DE item_info *)
    mo_provides : item_info srcflaged list option;
    mo_body : packbody ;
  }
and pack_info = {
    pa_name : Ident.pack_name ;
    pa_def : pack_def ;
  }
and
  pack_def = 
    PackGiven of pack_given
  | PackInstance of pack_instance
and
  pack_given = { 
    pg_uses  : Ident.pack_name srcflaged list ;
    (* N.B. CAS PARTICULIER DE item_info *)
    pg_provides : item_info srcflaged list option;
    pg_body : packbody ;
  }
and
  pack_instance = {
    pi_model : Ident.t ;
    pi_args : static_arg srcflaged list ;
  }
    (** Collection de noeuds, types const etc.
	- une table pour chaque sorte de defs
	- une liste de defs permettant de les
	ressortir dans l'ordre                   
    *)
and
  packbody = {
    pk_const_table : (Ident.t, const_info srcflaged ) Hashtbl.t ;
    pk_type_table  : (Ident.t, type_info  srcflaged ) Hashtbl.t ;
    pk_oper_table  : (Ident.t, oper_info  srcflaged ) Hashtbl.t ;
    pk_def_list    : item_ident list ; 
  }
and
  item_ident
  =
    (** La liste des items permet (uniquement ?)  de retrouver
	l'ordre de dfinitions *)
    ConstItem of Ident.t
  | TypeItem  of Ident.t
  | OperItem  of Ident.t
and
  item_info =
    ConstInfo of const_info
  | TypeInfo  of type_info
  | OperInfo  of oper_info
and
  (**************************************************)
  (** Info associe  un identificateur de constante *)
  (**************************************************)
  const_info = 
    ExternalConst  of (Ident.t * type_exp)
  | EnumConst      of (Ident.t * type_exp)
  | DefinedConst   of (Ident.t * type_exp option * val_exp)
and
  (**************************************************)
  (** Info associe  un identificateur de type *)
  (**************************************************)
  type_info =
    ExternalType of (Ident.t)
  | AliasedType  of (Ident.t * type_exp)
  | EnumType     of (Ident.t * Ident.t srcflaged list)
	  (* Pour les struct, on a un type un peu plus complexe *)
  | StructType   of struct_type_info
and
  (***************************************************)
  (* Type structure                                  *)
  (***************************************************)
  (** Info de type structure                         *)
  (* On garde :                                      *)
  (* - la liste des noms de champ                    *)
  (* - une table d'info de champ                     *)
  (***************************************************)
  struct_type_info = {
    st_name : Ident.t ;
    st_flist : Ident.t list ;
    st_ftable  : (Ident.t, field_info srcflaged)  Hashtbl.t 
  }
and field_info = {
  fd_name  : Ident.t ;
  fd_type  : type_exp ;
  fd_value : val_exp option
}
and
  (************************************************************)
  (** Type "immdiat" : type les flots, params, constantes etc *) 
  (* N.B. : toujours srcflaged, soit par le lexeme de l'ident *)
  (* soit sur l'oprateur ^ des tableaux                      *) 
  (************************************************************)
  type_exp = type_exp_core srcflaged
and
  type_exp_core =
    Bool_type_exp
  | Int_type_exp
  | Real_type_exp
  | Named_type_exp  of Ident.idref 
  | Array_type_exp  of (type_exp * val_exp)
and
  (*************************************************)
  (** Info associe  un identificateur d'operateur *)
  (** - Soit un node, soit une fonction externe     *)
  (*************************************************)
  oper_info =
    Node  of node_info
  | Func  of func_info
and
  (*************************************************)
  (** Info associe  un identificateur de fonction *)
  (*************************************************)
  func_info = {
    fn_name    : Ident.t ;
    fn_inpars  : (Ident.t option * type_exp) list ;
    fn_outpars : (Ident.t option * type_exp) list
  }
and
  (**********************************************)
  (** Info associe  un identificateur de noeud *)
  (** Soit une def complete, soit un alias       *)
  (**********************************************)
  node_info = {
    ni_name       : Ident.t ;
    ni_staticlist : static_param srcflaged list ;
    ni_value : node_def
  }
and node_profile = (var_info srcflaged list * var_info srcflaged list)
and node_def =
    NodeExtern of node_profile
  | NodeBody  of node_body
  | NodeAlias of node_profile option * oper_exp srcflaged
	  (**********************************************)
	  (* Pour une def complete, on garde :          *)
	  (* - une table des variables avec src_info    *)
	  (* - une liste des noms de vars pour chaque   *)
	  (* sorte de var                               *)
	  (**********************************************)
and node_body = {
  nbdy_inlist  :   Ident.t list ;
  nbdy_outlist :   Ident.t list ;
  nbdy_loclist :   Ident.t list ;
  nbdy_vartable:   var_info_table ;
  nbdy_asserts :   (val_exp srcflaged) list ;
  nbdy_eqs     :   (eq_info srcflaged) list
}
and var_info_table = (Ident.t, var_info srcflaged)  Hashtbl.t
and static_param =
    StaticParamType  of Ident.t
  | StaticParamConst of (Ident.t * type_exp)
  | StaticParamNode  of (Ident.t * var_info srcflaged list * var_info srcflaged list)
and var_info = {
  va_nature : var_nature;
  va_name   : Ident.t;
  va_type   : type_exp;
  va_clock  : clock_exp 
}
and var_nature =
    VarInput
  | VarOutput
  | VarLocal
and eq_info = {
  eq_left : left_part list ;
  eq_right: val_exp
}
    (* Information d'horloge pour les variables de noeud *)
and clock_exp =
    BaseClock
  | NamedClock of Ident.t srcflaged
and left_part = 
    (* tous les acces en partie gauche sont flags *)
    LeftVar of (Ident.t srcflaged)
  | LeftField of (left_part * (Ident.t srcflaged))
  | LeftArray of (left_part * (val_exp srcflaged))  
  | LeftSlice of (left_part * (slice_info srcflaged))
and slice_info = {
  si_first : val_exp ;
  si_last  : val_exp ;
  si_step  : val_exp option ;
}

(************************************************)
(* Info associes aux expressions               *)
(************************************************)
(* Vision "fonctionnelle" des val_exp :         *)
(* Une exp. est une application d'operation :   *)
(* - avec passage par position, auquel cas les  *)
(* oprandes sont des val_exp                   *)
(* - avec passage par nom, auquel cas les       *)
(* oprandes sont des Ident.t * val_exp          *)
(************************************************)
and val_exp =
    CallByPos  of (by_pos_op  srcflaged  * val_exp list) 
  | CallByName of (by_name_op srcflaged  * (Ident.t srcflaged * val_exp) list)
   
	  (* Toutes les oprations classiques *)
and by_pos_op =
    (* zeroaire *)
    NULL_exp
  | TRUE_n
  | FALSE_n
  | ICONST_n of Ident.t
  | RCONST_n of Ident.t
  | IDENT_n  of Ident.idref
	  (* unaires *)
  | NOT_n
  | UMINUS_n
  | PRE_n
  | CURRENT_n
  | REAL2INT_n
  | INT2REAL_n
      (* binaires *)
  | ARROW_n
  | FBY_n
  | WHEN_n
  | AND_n
  | OR_n
  | XOR_n
  | IMPL_n
  | EQ_n
  | NEQ_n
  | LT_n
  | LTE_n
  | GT_n
  | GTE_n
  | DIV_n
  | MOD_n
  | MINUS_n
  | PLUS_n
  | SLASH_n
  | TIMES_n
  | POWER_n
  | HAT_n
  | CONCAT_n
      (* ternaires *)
  | IF_n
  | WITH_n
      (* n-aires *)
  | NOR_n
  | DIESE_n
  | TUPLE_n
  | ARRAY_n
      (************************************************)
      (* Info associes aux expressions  (suite)      *)
      (************************************************)
      (* pseudo-unaire : A USAGE INTERNE *)
      (* projection de tuple             *)
  | PROJ_n of int
      (* pseudo-unaire : appel par position *)
  | CALL_n of oper_exp srcflaged
      (* pseudo-unaire : acces tableau *)
  | ARRAY_ACCES_n of val_exp
  | ARRAY_SLICE_n of slice_info
      (* pseudo-unaire : acces structure *)
  | STRUCT_ACCESS_n    of Ident.t
      (* Opration avec passage de param par nom *)

(*   | MERGE_n of Ident.t * ((Ident.t * val_exp) list) *)
      

and by_name_op =
    BYNAME_CALL_n of Ident.idref
      (* structure non type *)
  | STRUCT_EXP_n
      (* Descripteur d'operateur call *)
and oper_exp =
    CallPreDef of by_pos_op
  | CallUsrDef of (Ident.idref * static_arg srcflaged list)
	  (*
	    Params statiques effectifs :
	    - val_exp (pour les constantes)
	    - type_exp (pour les types)
	    - oper_exp (pour les node)
	    - ident : a rsoudre, peut etre const, type ou node 
	  *)
and static_arg =
    StaticArgIdent of Ident.idref
  | StaticArgConst of val_exp
  | StaticArgType  of type_exp
  | StaticArgNode  of oper_exp



(**********************************************
Utilitaires pour fabriquer des packages
**********************************************)
let give_pack_this_name nme pbdy = (
   {
      pa_name = nme;
      pa_def = PackGiven {
      	pg_uses = [];
      	pg_provides = None;
      	pg_body = pbdy;
		}
   }
)

(**********************************************
Utilitaires pour fabriquer des val_exp
**********************************************)

let leafexp lxm op = (
  CallByPos ( {src = lxm ; it = op } , [] )	
) 

let unexp lxm op e1 = (
  CallByPos ( {src = lxm ; it = op } , [e1] )	
) 

let binexp lxm op e1 e2 = (
  CallByPos ( {src = lxm ; it = op } , [e1 ; e2] )	
) 

let ternexp lxm op e1 e2 e3 = (
  CallByPos ( {src = lxm ; it = op } , [e1 ; e2; e3] )	
) 

let naryexp lxm op elst = (
  CallByPos ( {src = lxm ; it = op } , elst )	
) 

let bynameexp lxm op nelst = (
  CallByName ( {src = lxm ; it = op } , nelst )	
) 

(*----------------------------------------------------------------------------*)
(*                      INTERFACE AVEC LE PARSER                              *)
(*----------------------------------------------------------------------------*)


(**********************************************
Construction d'un packbody
n.b. les tables sont copies et donc
rutilisables par l'appelant
**********************************************)

let make_packbody ctab ttab otab dlst = (
  {
    pk_const_table = Hashtbl.copy ctab;
    pk_type_table  = Hashtbl.copy ttab;
    pk_oper_table  = Hashtbl.copy otab;
    pk_def_list    = dlst
  }
)

(*---------------------------------------------------------------------
add_info
-----------------------------------------------------------------------
Rle :
	proc gnrique pour mettre une info 'a dans
	 une table (Ident.t, 'a srcflaged).

Effets de bord :
	erreur de compil si dj utilis 	
----------------------------------------------------------------------*)
let add_info 
    (htbl   : (Ident.t, 'a srcflaged) Hashtbl.t)
    (kindof : string)   (* une string en cas d'erreur   *)
    (lxm  : Lxm.t)      (* le lexeme en question        *)
    (info : 'a)         (* l'info en question           *)
    = (
      try
	let x  = Hashtbl.find htbl (Lxm.id lxm) in 
	  raise (
	    Compile_error ( 
	      lxm ,
	      (sprintf "bad %s declaration, ident already linked at %s"
		 kindof (Lxm.position x.src)
	      )
	    )
	  )
      with Not_found -> (
	Hashtbl.add htbl (Lxm.id lxm) { src = lxm ; it  = info }
      )
    )
(*---------------------------------------------------------------------
lexeme_of_val_exp
lexeme_of_left_part
-----------------------------------------------------------------------
Rle : retourne le lexeme ``principal'' d'une expression/d'un 

Entres : val_exp

Sorties : Lxm.t

Effets de bord :
----------------------------------------------------------------------*)
let lexeme_of_val_exp = function
    CallByPos  (opflaged, args) -> opflaged.src
  | CallByName (opflaged, args) -> opflaged.src

let rec lexeme_of_left_part = function
  | LeftVar sflg -> sflg.src
  | LeftField (x, _) -> lexeme_of_left_part x 
  | LeftArray (x, _) -> lexeme_of_left_part x
  | LeftSlice (x, _) -> lexeme_of_left_part x
	  

let lexeme_of_pack_or_model = (
  function 
      NSPack p -> p.src
    | NSModel m -> m.src
)

(********************************************)
let (pack_or_model_to_string: pack_or_model -> string) =
  function
    | NSPack  pi -> Ident.pack_name_to_string pi.it.pa_name ^ " (pack) "
    | NSModel mi -> Ident.pack_name_to_string mi.it.mo_name ^ " (model) "
