(** Time-stamp: <modified the 19/12/2007 (at 11:05) by Erwan Jahier> *)


open Lxm
open SyntaxTree
open Format

(***********************************************************************************)
(* exported *)

let (op2string : SyntaxTree.by_pos_op -> string) =
  fun op ->  
  match op with 
      (* unaires *)
    | (NOT_n    ) -> "not"
    | (UMINUS_n ) -> "-"
    | (PRE_n    ) -> "pre"
    | (CURRENT_n) -> "current"
	(* binaires *)
    | (ARROW_n  ) -> "->"
    | (WHEN_n   ) -> "when"
    | (AND_n    ) -> "and"
    | (OR_n     ) -> "or"
    | (XOR_n    ) -> "xor"
    | (IMPL_n   ) -> "=>"
    | (EQ_n     ) -> "="
    | (NEQ_n    ) -> "<>"
    | (LT_n     ) -> "<"
    | (LTE_n    ) -> "<="
    | (GT_n     ) -> ">"
    | (GTE_n    ) -> ">="
    | (DIV_n    ) -> "div"
    | (MOD_n    ) -> "mod"
    | (MINUS_n  ) -> "-"
    | (PLUS_n   ) -> "+"
    | (SLASH_n  ) -> "/"
    | (TIMES_n  ) -> "*"
    | (POWER_n  ) -> "**"
    | (HAT_n    ) -> "^"
    | (CONCAT_n ) -> "|"
    | (PROJ_n _ ) -> assert false
    | (IDENT_n _) -> assert false
    | (RCONST_n _)-> assert false
    | (ICONST_n _)-> assert false
    | (FBY_n    ) -> assert false
    | (INT2REAL_n)-> assert false
    | (REAL2INT_n)-> assert false
    | (FALSE_n  ) -> assert false
    | (TRUE_n   ) -> assert false
    | (NULL_exp ) -> assert false
	(* ternaires *)
    | (IF_n     ) -> "if"
    | (WITH_n   ) -> "with" 
	(* n-aires *)
    | (NOR_n    ) -> "nor"
    | (DIESE_n  ) -> "#"
    | (TUPLE_n  ) -> assert false
    | (CALL_n _ ) -> assert false
	(* tableau (spcial) *)
    | (ARRAY_n  ) -> assert false
	(* pseudo-unaire : acces tableau *)
    | (ARRAY_ACCES_n _ ) ->  assert false
    | (ARRAY_SLICE_n sl) -> assert false
	(* pseudo-unaire : acces structure *)
    | (STRUCT_ACCESS_n fld) -> assert false


(***********************************************************************************)
(* exported *)
let rec packbody (os: Format.formatter) (pkg: SyntaxTree.packbody) = 
  let dump_def (d: item_ident) = (
    (
      try ( 
	match d with
	    ConstItem id -> dump_const os (Hashtbl.find pkg.pk_const_table id)
	  | TypeItem  id -> dump_type os (Hashtbl.find pkg.pk_type_table id)
	  | OperItem  id -> (
	      match (Hashtbl.find pkg.pk_oper_table id) with
		| {src = lxm ; it = Node (ninfo) } -> 
		    dump_node os {src = lxm; it = ninfo}
		| {src = lxm ; it = Func finfo } ->
		    dump_func os {src = lxm; it = finfo}
	    )
      ) 
      with Not_found -> 
	print_string ("*** unable to find a definition for " ^
			(match d with 
			     ConstItem id 
			   | TypeItem  id 
			   | OperItem  id -> Ident.to_string id
			));
	assert false
    ) ;
    Format.fprintf os "@\n"
  ) in 
    (* Format.fprintf os "@?@[<b 0>" ; *)
    List.iter dump_def pkg.pk_def_list  ;
    (* Format.fprintf os "@]@."  *)

(*******************************)
(* dump d'une def de constante *)
(*******************************)
and dump_const (os: Format.formatter) (x: const_info srcflaged) = (
  let lxm = x.src and info = x.it in
    fprintf os "-- %s@\n" (Lxm.details lxm) ;
    fprintf os "const %s " (Lxm.str lxm) ;
    dump_const_def os info ;
    fprintf os " ;@\n" ;
)
and dump_const_def (os: Format.formatter) (info: const_info) = (
  match info with
    | ExternalConst (id, ty) -> fprintf os ": %a " dump_type_exp ty
    | DefinedConst  (id, None, exp) -> fprintf os "= %a " dump_val_exp exp
    | DefinedConst  (id, Some ty, exp) -> 
	fprintf os ": %a = %a " dump_type_exp ty dump_val_exp exp
    | EnumConst  (id, ty) -> fprintf os "XXX printme!"
)
  (**************************)
  (* dump d'une def de type *)
  (**************************)
and dump_type (os: Format.formatter) (x: type_info srcflaged) = (
  let lxm = x.src and info = x.it in
  let id = Lxm.str lxm in
    fprintf os "-- %s@\n" (Lxm.details lxm) ;
    fprintf os "type %s " id ;
    dump_type_def os info ;
    fprintf os " ;@\n" ;
)
and dump_type_def (os: Format.formatter) (info: type_info ) = (
  match info with
    | ExternalType id -> () 
    | AliasedType (id, te) -> fprintf os "= %a" dump_type_exp te
    | EnumType (id, clist) ->
	(* fprintf os "= enum {@[<b 3>@,%a@,@]}" dump_id_list clist *)
	fprintf os "= enum { %a}" dump_id_list clist
       
    | StructType {st_name = id; st_flist = fl; st_ftable = ft} -> (
	let filst = List.map (function id -> (Hashtbl.find ft id).it) fl in
	  fprintf os "= struct {@\n@[<b 3>   %a@]@\n}"
	    dump_field_list filst
      )
)
  (****************************)
  (* dump d'une liste de noms *)
  (****************************)
and dump_id_list (os : formatter) (idlst : Ident.t srcflaged list) = (
  match idlst with 
      [] -> () 
    | h::[] -> ( fprintf os "%s" (Ident.to_string h.it))
    | h::t  -> ( fprintf os "%s, %a" (Ident.to_string h.it) dump_id_list t)
)
  (*****************************)
  (* dump d'une liste de field *)
  (*****************************)
and dump_field_list (os: Format.formatter) (filst: field_info list) = (
  match filst with
      [] -> ()
    | h::[] -> ( fprintf os "%a" dump_field h )
    | h::t	-> ( fprintf os "%a;@\n%a" dump_field h dump_field_list t)
)
and dump_field (os: Format.formatter) (finfo: field_info) = (
  match finfo with
      {fd_name=id; fd_type=ty; fd_value=None} -> (
	fprintf os "%s : %a" (Ident.to_string id) dump_type_exp ty
      ) 
    |
	{fd_name=id; fd_type=ty; fd_value=Some ex} -> (
	  fprintf os "%s : %a = %a" (Ident.to_string id) 
	    dump_type_exp ty dump_val_exp ex
	)
)
  (**************************)
  (* dump d'une def de func *)
  (**************************)
and dump_func (os: Format.formatter) (x: func_info srcflaged) = (
  let lxm = x.src and info = x.it in
    fprintf os "-- %s" (Lxm.details lxm) ;
    fprintf os " (external function)@\n" ;
    fprintf os "function %s (@,@[<b 3>%a@]@,) returns (@,@[<b 3>%a@]@,);@\n"
      (Ident.to_string info.fn_name)
      dump_param_list info.fn_inpars
      dump_param_list info.fn_outpars
)
and dump_param_list 
    (os: Format.formatter)
    (plst: (Ident.t option * type_exp) list) = 
  (
    match plst with
	[] -> ()
      | h::[] -> ( fprintf os "%a" dump_param h )
      | h::t  -> ( fprintf os "%a; %a" dump_param h dump_param_list t) 
  )
and dump_param (os: Format.formatter) (p: (Ident.t option * type_exp)) = (
  match p with
      (None, ty) -> (fprintf os "%a" dump_type_exp ty)
    | (Some id, ty) -> (fprintf os "%s : %a" (Ident.to_string id) dump_type_exp ty) 
) 
  (**************************)
  (* dump d'une eq. de node *)
  (**************************)
and dump_item_info_list 
    (os: Format.formatter) 
    (lst: item_info srcflaged list) =
  (
    match lst with
	[] -> ()
      | h::[]	-> dump_item_info os h
      | h::t  -> fprintf os "%a;@\n%a" dump_item_info h dump_item_info_list t 
  )
and dump_item_info
    (os: Format.formatter)
    (ie: item_info srcflaged) =
  (
    match ie.it with
      | ConstInfo ci -> dump_const os (Lxm.flagit ci ie.src)
      | TypeInfo ti -> dump_type os (Lxm.flagit ti ie.src)
      | OperInfo oi -> (
	  match oi with
	    | Node ninfo -> dump_node os (Lxm.flagit ninfo ie.src)
	    | Func finfo -> dump_func os (Lxm.flagit finfo ie.src)
	)
  )

and dump_static_param_list 
    (os: Format.formatter) 
    (lst: static_param srcflaged list) =
  (
    match lst with
	[] -> ()
      | h::[]	-> dump_static_param os h
      | h::t  -> fprintf os "%a;@\n%a" dump_static_param h dump_static_param_list t 
  )
and dump_static_param 
    (os: Format.formatter) 
    (sp: static_param srcflaged) =
  (
    match sp.it with
      | StaticParamType id ->  fprintf os "type %s" (Ident.to_string id)
      | StaticParamConst (id, exp) -> fprintf os "const %s : %a" 
	  (Ident.to_string id) dump_type_exp exp
      | StaticParamNode (id, ins, outs) -> (
	  fprintf os "node %s(@,%a@,)returns(@,%a@,)" (Ident.to_string id)
	    dump_line_var_decl_list ins dump_line_var_decl_list outs
	)
  )
    (**************************)
    (* dump d'une def de node *)
    (**************************)
and dump_node (os: Format.formatter) (x: node_info srcflaged) = (
  let lxm = x.src and ninfo = x.it in
    fprintf os "-- %s" (Lxm.details lxm) ;
    fprintf os " (node definition)@\n" ;
    fprintf os "node %s" (Ident.to_string ninfo.ni_name);
    if (ninfo.ni_staticlist != []) then (
      fprintf os " <<@\n" ;
      fprintf os "@[<b 3>@   %a@]@\n" dump_static_param_list ninfo.ni_staticlist ;
      fprintf os ">>\n"
    ) else () ;	
    match ninfo.ni_value with
	NodeBody nbody -> dump_node_body os nbody
      | NodeAlias (_,{it = nexp; src = lxm}) -> 
	  fprintf os " = @,%a;@\n" dump_oper_exp nexp
      | NodeExtern _ -> assert false
)

and dump_node_body (os: Format.formatter) (nbody: node_body) = (
  let vartab = nbody.nbdy_vartable in
  let get_info (id: Ident.t) = (
    (Hashtbl.find vartab id).it
  ) in
  let inlst = List.map get_info nbody.nbdy_inlist in
  let outlst = List.map get_info nbody.nbdy_outlist in
  let loclst = List.map get_info nbody.nbdy_loclist in
    fprintf os "(@\n" ;
    fprintf os "@[<b 3>@   %a@]@\n" dump_var_decl_list inlst ;
    fprintf os ") returns (@\n" ;
    fprintf os "@[<b 3>@   %a@]@\n" dump_var_decl_list outlst ;
    fprintf os ");@\n" ;
    if (nbody.nbdy_loclist != []) then (
      fprintf os "var@\n";
      fprintf os "@[<b 3>   %a;@]@\n" dump_var_decl_list loclst ;
    ) else ();
    (* le corps *)
    fprintf os "@[<b 3>let" ;
    (* les assertions *)
    dump_assert_list os nbody.nbdy_asserts ;	
    (* les equations *)
    dump_eq_list os nbody.nbdy_eqs ;
    fprintf os "@]@\ntel" ;
)
  (* dclarations sur plusieurs lignes, indentes ... *)
and dump_var_decl_list (os: Format.formatter) (lst: var_info list) = (
  match lst with
      [] -> ()
    | h::[] -> ( fprintf os "%a" dump_var_decl h )
    | h::t -> ( fprintf os "%a;@\n%a" dump_var_decl h dump_var_decl_list t )
)
  (* dclarations sur une ligne AVEC SOURCE ... *)
and dump_line_var_decl_list (os: Format.formatter) (lst: var_info srcflaged list) = (
  match lst with
      [] -> ()
    | h::[] -> ( fprintf os "%a" dump_var_decl h.it )
    | h::t -> ( fprintf os "%a;@,%a" dump_var_decl h.it dump_line_var_decl_list t )
)
and dump_var_decl (os: Format.formatter) (vinfo: var_info  ) = (
  fprintf os "%s : %a" (Ident.to_string vinfo.va_name) dump_type_exp vinfo.va_type ;
  (
    match vinfo.va_clock with
	BaseClock -> ()
      | NamedClock {it=id;src=lxm} -> (fprintf os " when %s" (Ident.to_string id))
  )
)
  (**************************)
  (* dump d'un type immdiat*)
  (**************************)
and dump_type_exp (os: Format.formatter) (x: type_exp) = (
  match x.it with
      Bool_type_exp -> fprintf os "bool"
    |  Int_type_exp  -> fprintf os "int"
    |  Real_type_exp -> fprintf os "real"
    |  Named_type_exp id -> fprintf os "%s" (Ident.string_of_idref id)
    |  Array_type_exp (te, sz) -> (
	 dump_type_exp os te ;
	 fprintf os "^" ;
	 dump_val_exp os sz
       )
)
  (**************************)
  (* dump des assertions    *)
  (**************************)
and dump_assert_list (os: Format.formatter) (af: (val_exp srcflaged) list) = (
  match af with
      [] -> ()
    | a :: reste -> (
	  Format.fprintf os "@\nassert %a;" dump_val_exp a.it ;
	  Format.fprintf os "%a" dump_assert_list reste
	)
)
  (**************************)
  (* dump des equations     *)
  (**************************)
and dump_eq_list (os: Format.formatter) (eqfs: (eq_info srcflaged) list) = (
  match eqfs with
      [] -> ()
    | {it={eq_left=lflst; eq_right=exp}; src=lxm} :: reste -> (
	  Format.fprintf os "@\n%a = %a;%a"
	    dump_left_part_list lflst
	    dump_val_exp exp
	    dump_eq_list reste
	)
)
and dump_left_part_list (os: Format.formatter) (lfts: left_part list) =
  (
    match lfts with
	l::[] ->     dump_left_part os l
      | l::reste ->  fprintf os "%a,@,%a" dump_left_part l dump_left_part_list reste
      | _ -> assert false
	  
  )
and dump_left_part (os: Format.formatter) (lft: left_part) =
  (
    match lft with
	LeftVar idflg ->  fprintf os "%s" (Ident.to_string idflg.it)
      | LeftField (l, idflg) ->  fprintf os "%a.%s" dump_left_part l (Ident.to_string idflg.it)
      | LeftArray (l, expflg) ->
	  fprintf os "%a[@,%a@,]" dump_left_part l dump_val_exp expflg.it
      | LeftSlice (l, slcflg) ->
	  fprintf os "%a[@,%a@,]" dump_left_part l dump_slice_info slcflg.it
  )
    (**************************)
    (* dump d'une expression  *)
    (**************************)
and dump_val_exp (os: Format.formatter) (x: val_exp) = (
  match x with
      CallByPos ( {it=oper; src=lxm} , pars ) -> (
	dump_by_pos_exp os oper pars
      ) |
	  CallByName ( {it=oper; src=lxm}, nm_pars ) -> (
	    dump_by_name_exp os oper nm_pars
	  )
)
and dump_val_exp_list (os : formatter) (xl: val_exp list) = (
  match xl with 
      [] -> () 
    | h::[] -> ( fprintf os "%a" dump_val_exp h ) 
    | h::t  -> ( fprintf os "%a,@,%a" dump_val_exp h dump_val_exp_list t)
)
  (****************************)
  (* Appels "par position"    *)
  (****************************)
and dump_by_pos_exp (os: Format.formatter) (oper: by_pos_op) (pars: val_exp list) =
  (
    match (oper, pars) with 
	(TRUE_n,     [])  -> dump_leaf_exp os "true" 
      | (FALSE_n,    [])  -> dump_leaf_exp os "false"
      | (ICONST_n s, [])  -> dump_leaf_exp os (Ident.to_string s) 
      | (RCONST_n s, [])  -> dump_leaf_exp os (Ident.to_string s)
      | (IDENT_n  id, [])  -> dump_leaf_exp os (Ident.string_of_idref id)
	    (* unaires *)
      | (NOT_n,       [p0]) -> dump_unary_exp os "not" p0
      | (UMINUS_n,    [p0]) -> dump_unary_exp os "-" p0
      | (PRE_n,       [p0]) -> dump_unary_exp os "pre" p0
      | (CURRENT_n,   [p0]) -> dump_unary_exp os "current" p0
      | (REAL2INT_n,  [p0]) -> dump_unary_exp os "int" p0
      | (INT2REAL_n,  [p0]) -> dump_unary_exp os "real" p0
	  (* binaires *)
      | (ARROW_n,  [p0;p1]) -> dump_binary_exp os "->" p0 p1
      | (FBY_n,    [p0;p1]) -> dump_binary_exp os "fby" p0 p1
      | (WHEN_n,   [p0;p1]) -> dump_binary_exp os "when" p0 p1
      | (AND_n,    [p0;p1]) -> dump_binary_exp os "and" p0 p1
      | (OR_n,     [p0;p1]) -> dump_binary_exp os "or" p0 p1
      | (XOR_n,    [p0;p1]) -> dump_binary_exp os "xor" p0 p1
      | (IMPL_n,   [p0;p1]) -> dump_binary_exp os "=>" p0 p1
      | (EQ_n,     [p0;p1]) -> dump_binary_exp os "=" p0 p1
      | (NEQ_n,    [p0;p1]) -> dump_binary_exp os "<>" p0 p1
      | (LT_n,     [p0;p1]) -> dump_binary_exp os "<" p0 p1
      | (LTE_n,    [p0;p1]) -> dump_binary_exp os "<=" p0 p1
      | (GT_n,     [p0;p1]) -> dump_binary_exp os ">" p0 p1
      | (GTE_n,    [p0;p1]) -> dump_binary_exp os ">=" p0 p1
      | (DIV_n,    [p0;p1]) -> dump_binary_exp os "div" p0 p1
      | (MOD_n,    [p0;p1]) -> dump_binary_exp os "mod" p0 p1
      | (MINUS_n,  [p0;p1]) -> dump_binary_exp os "-" p0 p1
      | (PLUS_n,   [p0;p1]) -> dump_binary_exp os "+" p0 p1
      | (SLASH_n,  [p0;p1]) -> dump_binary_exp os "/" p0 p1
      | (TIMES_n,  [p0;p1]) -> dump_binary_exp os "*" p0 p1
      | (POWER_n,  [p0;p1]) -> dump_binary_exp os "**" p0 p1
      | (HAT_n,    [p0;p1]) -> dump_binary_exp os "^" p0 p1
      | (CONCAT_n, [p0;p1]) -> dump_binary_exp os "|" p0 p1
	  (* ternaires *)
      | (IF_n,   [p0;p1;p2]) -> dump_ternary_exp os "if" "then" "else" p0 p1 p2
      | (WITH_n, [p0;p1;p2]) -> dump_ternary_exp os "with" "then" "else" p0 p1 p2
	  (* n-aires *)
      | (NOR_n,    pl) -> dump_nary_exp os "nor" pl
      | (DIESE_n,  pl) -> dump_nary_exp os "#" pl
      | (TUPLE_n,  pl) -> dump_nary_exp os "" pl
      | (CALL_n s, pl) -> fprintf os "%a(@,%a@,)"
	  dump_oper_exp s.it dump_val_exp_list pl 
	    (* tableau (spcial) *)
      | (ARRAY_n,  pl) -> fprintf os "[@,%a@,]" dump_val_exp_list pl
	  (* pseudo-unaire : acces tableau *)
      | (ARRAY_ACCES_n ix, [p0]) ->  fprintf os "%a[@,%a@,]" 
	  dump_val_exp p0 dump_val_exp ix 
      | (ARRAY_SLICE_n sl, [p0]) ->  fprintf os "%a[@,%a@,]" 
	  dump_val_exp p0 dump_slice_info sl 
	    (* pseudo-unaire : acces structure *)
      | (STRUCT_ACCESS_n fld, [p0]) -> fprintf os "%a.%s"
	  dump_val_exp p0 (Ident.to_string fld)

	    
      | (FBY_n, _) -> assert false
      | (PROJ_n _, _) -> assert false
      | (NULL_exp, _) -> assert false

      | (STRUCT_ACCESS_n _, _) -> assert false
      | (ARRAY_SLICE_n _, _) -> assert false
      | (ARRAY_ACCES_n _, _) -> assert false
      | (WITH_n, _) -> assert false
      | (IF_n, _) -> assert false
      | (CONCAT_n, _) -> assert false
      | (HAT_n, _) -> assert false
      | (POWER_n, _) -> assert false
      | (TIMES_n, _) -> assert false
      | (SLASH_n, _) -> assert false
      | (PLUS_n, _) -> assert false
      | (MINUS_n, _) -> assert false
      | (MOD_n, _) -> assert false
      | (DIV_n, _) -> assert false
      | (GTE_n, _) -> assert false
      | (GT_n, _) -> assert false
      | (LTE_n, _) -> assert false
      | (LT_n, _) -> assert false
      | (NEQ_n, _) -> assert false
      | (EQ_n, _) -> assert false
      | (IMPL_n, _) -> assert false
      | (XOR_n, _) -> assert false
      | (OR_n, _) -> assert false
      | (AND_n, _) -> assert false
      | (WHEN_n, _) -> assert false
      | (ARROW_n, _) -> assert false
      | (INT2REAL_n, _) -> assert false
      | (REAL2INT_n, _) -> assert false
      | (CURRENT_n, _) -> assert false
      | (PRE_n, _) -> assert false
      | (UMINUS_n, _) -> assert false
      | (NOT_n, _) -> assert false
      | (IDENT_n _, _) -> assert false
      | (RCONST_n _, _::_) -> assert false
      | (ICONST_n _, _::_) -> assert false
      | (FALSE_n, _::_) -> assert false
      | (TRUE_n, _::_) -> assert false

  )
    (* les procs standard pour les operateurs predefs *)
and dump_leaf_exp (os : Format.formatter) (s : string) = (
  fprintf os "%s" s
)
and dump_unary_exp
    (os : Format.formatter)
    (s : string)
    (op0: val_exp)
    = (
      fprintf os "(@,%s %a@,)" s dump_val_exp op0
    )
and dump_binary_exp
    (os : Format.formatter)
    (s : string)
    (op0: val_exp)
    (op1: val_exp)
    = (
      fprintf os "(@,%a %s %a@,)" dump_val_exp op0 s dump_val_exp op1
    )
and dump_ternary_exp
    (os : Format.formatter)
    (s : string)
    (t : string)
    (e : string)
    (op0: val_exp)
    (op1: val_exp)
    (op2: val_exp)
    = (
      fprintf os "(@,%s %a %s %a %s %a@,)"
	s dump_val_exp op0 t dump_val_exp op1 e dump_val_exp op2
    )
and dump_nary_exp
    (os : Format.formatter)
    (s: string)
    (ops: val_exp list)
    = (
      fprintf os "%s(@,%a@,)" s dump_val_exp_list ops
    )

and dump_oper_exp
    (os : Format.formatter)
    (n: oper_exp)
    = (
      match n with
	  CallPreDef op -> (
	    fprintf os "operator %s" (op2string op)
	  ) |
	      (CallUsrDef (id, sal)) -> (
		fprintf os "%s" (Ident.string_of_idref id) ;
		( match sal with
		      [] -> ()
		    | lst -> (fprintf os "<< @,%a@, >>" dump_static_arg_list lst)
		)
	      )
    )
and dump_static_arg_list 
    (os : Format.formatter)
    (lst: static_arg srcflaged list)
    = (
      match lst with
	  [] -> ()
	| [sa] -> fprintf os "%a" dump_static_arg sa.it
	| sa::reste -> 
	    fprintf os "%a, @,%a" dump_static_arg sa.it dump_static_arg_list reste
    )
and dump_static_arg
    (os : Format.formatter)
    (sa: static_arg)
    = (
      match sa with
	  StaticArgIdent id -> (
	    fprintf os "%s" (Ident.string_of_idref id) 
	  ) |
	      StaticArgConst ve -> (
		fprintf os "const %a" dump_val_exp ve
	      ) |
		  StaticArgType  te -> (
		    fprintf os "type %a" dump_type_exp te 
		  ) |
		      StaticArgNode  ne -> (
			fprintf os "node %a" dump_oper_exp ne
		      ) 
    )
and dump_slice_info 
    (os : Format.formatter)
    (sl: slice_info)
    = (
      fprintf os "%a@, .. @,%a" dump_val_exp sl.si_first dump_val_exp sl.si_last ;
      match sl.si_step with
	  Some e -> fprintf os "@, step %a" dump_val_exp e
	| None -> ()
    )
  (****************************)
  (* Appels "par noms"        *) 
  (****************************)
and dump_by_name_exp
    (os: Format.formatter)
    (oper: by_name_op)
    (pars: (Ident.t srcflaged * val_exp) list) =
  (
    match (oper, pars) with
	(BYNAME_CALL_n id, pl) -> (
	  fprintf os "%s{@,%a@,}" 
	    (Ident.string_of_idref id) 
	    dump_named_pars pl
	) |
	    (STRUCT_EXP_n, pl) -> (
	      fprintf os "{@,%a@,}" dump_named_pars pl
	    )
  )
and dump_named_pars
    (os: Format.formatter)
    (pars: (Ident.t srcflaged * val_exp) list) =
  (
    match pars with
	[] -> (
	) |
	    (v,e)::[] -> (
	      fprintf os "%s = %a" (Ident.to_string v.it) dump_val_exp e
	    ) |
		(v,e)::l  -> (
		  fprintf os "%s = %a;@,%a" (Ident.to_string v.it) dump_val_exp e 
		    dump_named_pars l
		)
  )



let dump_packinstance (os: Format.formatter) (pi: SyntaxTree.pack_instance) = (
      Format.fprintf os "= %s(%a);@\n" 
	(Ident.to_string pi.pi_model)
	dump_static_arg_list pi.pi_args ;
    )

let dump_packgiven (os: Format.formatter) (pg: SyntaxTree.pack_given) = (
  (
    match (pg.pg_provides) with
	Some pl -> (
	  Format.fprintf os "provides@\n@[<b 3>   %a@]@,;@\n"
	    dump_item_info_list pl
	) | _ -> ()
  );
  Format.fprintf os "body@\n@[<b 3>   %a@]@\nend@\n"
    packbody  pg.pg_body ;
)

(***********************************************************************************)
(* exported *)
let rec packinfo (os: Format.formatter) (pf: SyntaxTree.pack_info srcflaged) = (
  let (p, lxm) = (pf.it, pf.src) in
    Format.fprintf os "@?@[<b 0>" ;
    Format.fprintf os "-----------------------------\n";
    Format.fprintf os "-- PACKAGE DEF \"%s\"\n" (Ident.pack_name_to_string p.pa_name);
    Format.fprintf os "-----------------------------\n";
    Format.fprintf os "-- %s\n" (Lxm.details lxm) ;
    (
      match (p.pa_def) with
	  PackGiven pg -> dump_packgiven os pg 
	| PackInstance pi -> dump_packinstance os pi 
    );
    Format.fprintf os "-----------------------------\n";
    Format.fprintf os "@]@." 
)
  

(***********************************************************************************)
(* exported *)
let modelinfo (os: Format.formatter) (mf: SyntaxTree.model_info srcflaged) = (
      let (m, lxm) = (mf.it, mf.src) in
	Format.fprintf os "@?@[<b 0>" ;
	Format.fprintf os "-----------------------------\n";
	Format.fprintf os "-- MODEL DEF \"%s\"\n" (Ident.pack_name_to_string m.mo_name);
	Format.fprintf os "-----------------------------\n";
	Format.fprintf os "-- %s\n" (Lxm.details lxm) ;
	Format.fprintf os "model %s@\n" (Ident.pack_name_to_string m.mo_name);

	( match (m.mo_needs) with
	      [] -> ()
	    | _ -> (
		Format.fprintf os "needs@\n@[<b 3>   %a@]@,;@\n"
		  dump_static_param_list m.mo_needs
	      )
	);
	( match (m.mo_provides) with
	      Some pl -> (
		Format.fprintf os "provides@\n@[<b 3>   %a@]@,;@\n"
		  dump_item_info_list pl
	      ) | _ -> ()
	);
	Format.fprintf os "body@\n@[<b 3>   %a@]@\nend@\n"
	  packbody  m.mo_body ;
	Format.fprintf os "-----------------------------\n";
	Format.fprintf os "@]@." 
    )

